--[[
    Aaron MacDonald
    August 18, 2010
]]

require('graphics')
require('constants')
require('actor')
require('levels')
require('collision')
require('ai')

local FPS = 60
local SECONDS_PER_FRAME = 1 / FPS

local accumulatedTime

local isPaused

----------------------------------------------------------------------

local KEY_QUIT = 'escape'

local VALID_BUTTONS = { [constants.UP] = true,
                        [constants.DOWN] = true,
                        [constants.LEFT] = true,
                        [constants.RIGHT] = true }

----------------------------------------------------------------------

local PLAYER_TYPE = {
    spriteSheet = 'player',
    spriteWidth = 64, spriteHeight = 96,
    spriteOffsetX = 32, spriteOffsetY = 64,
    radius = 32,
    speed = 5, diagonalSpeed = actor.diagonalSpeed(5)
}

local NPC_TYPE = {
    spriteSheet = 'npc',
    spriteWidth = 64, spriteHeight = 96,
    spriteOffsetX = 32, spriteOffsetY = 64,
    radius = 32,
    speed = 5, diagonalSpeed = actor.diagonalSpeed(5)
}

local PLAYER_START_FACING = constants.DOWN

----------------------------------------------------------------------

local currentButtons = { [constants.UP] = false,
                         [constants.DOWN] = false,
                         [constants.LEFT] = false,
                         [constants.RIGHT] = false }

local currentLevel
local allActors
local player

local camera

--local backgroundMusic

----------------------------------------------------------------------

local wanderAISteering = {
    [constants.NORTH] =
        function(movement)
            movement[constants.DOWN] = false
            movement[constants.LEFT] = false
            movement[constants.RIGHT] = false

            movement[constants.UP] = true
        end,
    [constants.NORTHEAST] =
        function(movement)
            movement[constants.DOWN] = false
            movement[constants.LEFT] = false

            movement[constants.UP] = true
            movement[constants.RIGHT] = true
        end,
    [constants.EAST] =
        function(movement)
            movement[constants.UP] = false
            movement[constants.DOWN] = false
            movement[constants.LEFT] = false

            movement[constants.RIGHT] = true
        end,
    [constants.SOUTHEAST] =
        function(movement)
            movement[constants.UP] = false
            movement[constants.LEFT] = false

            movement[constants.DOWN] = true
            movement[constants.RIGHT] = true
        end,
    [constants.SOUTH] =
        function(movement)
            movement[constants.UP] = false
            movement[constants.LEFT] = false
            movement[constants.RIGHT] = false

            movement[constants.DOWN] = true
        end,
    [constants.SOUTHWEST] =
        function(movement)
            movement[constants.UP] = false
            movement[constants.RIGHT] = false

            movement[constants.DOWN] = true
            movement[constants.LEFT] = true
        end,
    [constants.WEST] =
        function(movement)
            movement[constants.UP] = false
            movement[constants.DOWN] = false
            movement[constants.RIGHT] = false

            movement[constants.LEFT] = true
        end,
    [constants.NORTHWEST] =
        function(movement)
            movement[constants.DOWN] = false
            movement[constants.RIGHT] = false

            movement[constants.UP] = true
            movement[constants.LEFT] = true
        end,
}

local function wanderAISelectDir(brain)
    local i = math.random(1, #brain.possibleDirs)
    dir = brain.possibleDirs[i]
    wanderAISteering[dir](brain.movement)
end

local function wanderAIInit(brain)
    brain.possibleDirs = {constants.NORTH, constants.NORTHEAST,
      constants.EAST, constants.SOUTHEAST, constants.SOUTH,
      constants.SOUTHWEST, constants.WEST, constants.NORTHWEST}

    wanderAISelectDir(brain)

    brain.travelTime = 10
end

local function wanderAIThink(brain)
    if brain.travelTime == 0 then
        wanderAISelectDir(brain)
        brain.travelTime = 10
    else
        brain.travelTime = brain.travelTime - 1
    end
end

local function wanderHandleWallHit(brain)
    wanderAISelectDir(brain)
    brain.travelTime = 10
end

----------------------------------------------------------------------

local function spawnNPCs(level)
    local newNPCs = {}

    if level.npcSpawn then
        for _, data in ipairs(level.npcSpawn) do
            local npc = actor.newActor(NPC_TYPE, data.x, data.y,
              constants.DOWN)

            npc.ai = ai.newBrain(5, wanderAIThink, wanderAIInit,
              wanderHandleWallHit)

            table.insert(newNPCs, npc)
        end
    end

    return newNPCs
end

function love.load()
    math.randomseed( os.time() )

    accumulatedTime = 0
    isPaused = false

    graphics.loadFont()

    currentLevel = levels.newLevel('test_level')

    local startX = 0
    local startY = 0

    if currentLevel.playerSpawn then
        startX = currentLevel.playerSpawn.x
        startY = currentLevel.playerSpawn.y
    end

    allActors = {} --spawnNPCs(currentLevel)

    player = actor.newActor(PLAYER_TYPE, startX, startY,
      PLAYER_START_FACING)

    table.insert(allActors, player)

    local width = love.graphics.getWidth()
    local height = love.graphics.getHeight()

    camera = {
        x = 0,
        y = 0,
        minX = width / 2,
        minY = height / 2,
        maxX = levels.pixelWidth(currentLevel) - width,
        maxY = levels.pixelHeight(currentLevel) - height,
        width = width,
        height = height
    }
end

----------------------------------------------------------------------

function love.keypressed( key, unicode )
    if key == KEY_QUIT then
        love.event.push('q')
    elseif VALID_BUTTONS[key] then
        currentButtons[key] = true
    end
end

function love.keyreleased( key, unicode )
   if VALID_BUTTONS[key] then
       currentButtons[key] = false
   end
end

function love.focus(hasFocus)
    isPaused = not hasFocus
end

----------------------------------------------------------------------

local function moveActor(act)
    local frameTime = 1

    local moving = true
    local stoppedByWall = false

    while moving do
        local cData = collision.actorVSLevel(currentLevel, act,
          frameTime)

        if cData then
            assert(cData.time < frameTime)
            assert(cData.time >= 0)

            actor.move(act, cData.time)

            frameTime = frameTime - cData.time

            if frameTime > 0 then
                actor.setVelocity(act, cData.newDX, cData.newDY)
            else
                -- Stopped completely
                stoppedByWall = true
                moving = false
            end
        else
            -- No collision
            actor.move(act, frameTime)
            moving = false
        end
    end

    if stoppedByWall and act.ai then
        ai.wallHit(act.ai)
    end
end

local function moveAllActors()
    for _, act in ipairs(allActors) do
        if actor.isMoving(act) then
            moveActor(act)
        end
    end
end

local function updateCamera()
    if player.x < camera.minX then
        camera.x = 0
    elseif player.x > (camera.maxX + camera.minX) then
        camera.x = camera.maxX
    else
        camera.x = player.x - camera.minX
    end

    if (player.y - currentLevel.minY) < camera.minY then
        camera.y = currentLevel.minY
    elseif player.y > (camera.maxY + camera.minY) then
        camera.y = camera.maxY
    else
        camera.y = player.y - camera.minY
    end
end

local function myUpdate()
    -- Update velocities
    for _, act in ipairs(allActors) do
        local movementControls
        if act == player then
            movementControls = currentButtons
        else
            ai.think(act.ai)
            movementControls = act.ai.movement
        end

        actor.controlMovement(act, movementControls)

        if act ~= player then
            ai.update(act.ai)
        end
    end

    moveAllActors()
    updateCamera()
end

function love.update(dt)
    accumulatedTime = accumulatedTime + dt
    while accumulatedTime >= SECONDS_PER_FRAME
    do
        if not isPaused then
            myUpdate()
        end
        accumulatedTime = accumulatedTime - SECONDS_PER_FRAME
    end
end

----------------------------------------------------------------------

local function spriteCompare(a, b)
    local ay = a.y + (a.type.spriteHeight - a.type.spriteOffsetY)
    local by = b.y + (b.type.spriteHeight - b.type.spriteOffsetY)

    return ay < by
end

function love.draw()
    love.graphics.translate(-camera.x, -camera.y)

    levels.drawGround(currentLevel, camera)

    table.sort(allActors, spriteCompare)

    for _, act in ipairs(allActors) do
        actor.draw(act)
    end

    --levels.drawOver(currentLevel, camera)

    love.graphics.translate(camera.x, camera.y)

    love.graphics.print(player.x .. ', ' .. player.y, 0, 0)
end